@@ -1,10 +1,10 @@
 import json
 from datetime import timedelta
-from http.cookiejar import MozillaCookieJar
+from http.cookiejar import MozillaCookieJar, CookieJar
 from typing import Type, Optional, Union, List
 
 import requests.cookies
-from cachetools import TTLCache, cached
+from cachetools import TTLCache
 from .compat import BaseModel
 
 from fuo_bilibili.api.audio import AudioMixin
@@ -14,15 +14,12 @@ from fuo_bilibili.api.live import LiveMixin
 from fuo_bilibili.api.login import LoginMixin
 from fuo_bilibili.api.media import MediaMixin
 from fuo_bilibili.api.playlist import PlaylistMixin
-from fuo_bilibili.api.schema.enums import VideoQualityNum, SearchType
-from fuo_bilibili.api.schema.requests import BaseRequest, VideoInfoRequest, PlayUrlRequest, SearchRequest, \
-    FavoriteListRequest, PaginatedRequest, AudioFavoriteSongsRequest, AnotherPaginatedRequest, LivePlayUrlRequest, \
-    MediaFavlistRequest, HistoryAddLaterVideosRequest, BaseCsrfRequest, HistoryDelLaterVideosRequest, \
-    UserFollowingRequest, WeeklyDetailRequest
-from fuo_bilibili.api.schema.responses import BaseResponse
+from fuo_bilibili.api.schema.requests import *  # noqa
+from fuo_bilibili.api.schema.responses import *  # noqa
 from fuo_bilibili.api.user import UserMixin
 from fuo_bilibili.api.video import VideoMixin
-from fuo_bilibili.const import PLUGIN_API_COOKIEJAR_FILE, DANMAKU_DIRECTORY
+from fuo_bilibili.const import PLUGIN_API_COOKIEJAR_FILE
+from .wbi import encWbi
 
 CACHE = TTLCache(50, ttl=timedelta(minutes=10).total_seconds())
 
@@ -32,9 +29,19 @@ class BilibiliApi(BaseMixin, VideoMixin, LoginMixin, PlaylistMixin, HistoryMixin
     TIMEOUT = 10
 
     def __init__(self):
+        self._headers = {
+            'Referer': 'https://www.bilibili.com/',
+            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2)'
+                          ' AppleWebKit/537.36 (KHTML, like Gecko)'
+                          ' Chrome/33.0.1750.152 Safari/537.36'
+        }
         self._cookie = MozillaCookieJar(PLUGIN_API_COOKIEJAR_FILE)
         self._session = requests.Session()
+        # UPDATE(2024-01-11): To make self.nav_info work, the header needs to be added
+        # UPDATE(2023-xx-xx): The header cause some API failing: self.nav_info
+        self._session.headers.update(self._headers)
         self._session.cookies = self._cookie
+        self._wbi: Optional[NavInfoResponse.NavInfoResponseData.Wbi] = None
 
     @staticmethod
     def cookie_check():
@@ -42,17 +49,31 @@ class BilibiliApi(BaseMixin, VideoMixin, LoginMixin, PlaylistMixin, HistoryMixin
             return False
         return True
 
+    def get_session(self):
+        return self._session
+
+    def get_cookies(self):
+        return self._session.cookies
+
     @staticmethod
     def remove_cookie():
         PLUGIN_API_COOKIEJAR_FILE.unlink(missing_ok=True)
 
+    def from_cookiejar(self, jar: CookieJar):
+        for cookie in jar:
+            if 'bilibili.com' in cookie.domain:
+                self._cookie.set_cookie(cookie)
+        self._dump_cookie_to_file()
+
     def load_cookies(self):
         self._cookie.load()
 
     def _dump_cookie_to_file(self):
-        print('dumping cookies to file')
         self._cookie.save()
 
+    def set_wbi(self, wbi: NavInfoResponse.NavInfoResponseData.Wbi):
+        self._wbi = wbi
+
     @staticmethod
     def clear_cache_by_url(urls: List[str]):
         for key in CACHE.keys():
@@ -70,11 +91,24 @@ class BilibiliApi(BaseMixin, VideoMixin, LoginMixin, PlaylistMixin, HistoryMixin
 
     def get_uncached(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs) \
             -> Union[BaseResponse, BaseModel, None]:
-        print(f'Requesting: {url}...')
         if param is None:
             r = self._session.get(url, timeout=self.TIMEOUT, **kwargs)
         else:
-            r = self._session.get(url, timeout=self.TIMEOUT, params=json.loads(param.json(exclude_none=True)), **kwargs)
+            if isinstance(param, BaseCsrfRequest):
+                param.csrf = self._get_csrf()
+                js = json.loads(param.json(exclude_none=True))
+            elif isinstance(param, BaseWbiRequest):
+                if self._wbi is None:
+                    raise RuntimeError('wbi info is empty (not logged in)')
+                img_key = self._wbi.img_url.rsplit('/', 1)[1].split('.')[0]
+                sub_key = self._wbi.sub_url.rsplit('/', 1)[1].split('.')[0]
+                js = json.loads(param.json(exclude_none=True))
+                js = encWbi(js, img_key, sub_key)
+            else:
+                js = json.loads(param.json(exclude_none=True))
+            r = self._session.get(url, timeout=self.TIMEOUT,
+                                  params=js,
+                                  **kwargs)
         if r.status_code != 200:
             print(r.text)
             raise RuntimeError('http not 200')
@@ -83,19 +117,16 @@ class BilibiliApi(BaseMixin, VideoMixin, LoginMixin, PlaylistMixin, HistoryMixin
         response_str = r.text
         res: Union[BaseResponse, BaseModel] = clazz.parse_raw(response_str)
         if isinstance(res, BaseResponse) and res.code != 0:
-            raise RuntimeError(f'code not ok: {res.message}')
+            raise RuntimeError(f'code not ok: {res}, url:{url}, param:{param}')
         return res
 
-    @cached(CACHE)
     def get(self, url: str, param: Optional[BaseRequest], clazz: Union[Type[BaseResponse], Type[BaseModel], None], **kwargs)\
             -> Union[BaseResponse, BaseModel, None]:
         return self.get_uncached(url, param, clazz, **kwargs)
 
-    @cached(CACHE)
     def get_content(self, url: str) -> str:
         return self._session.get(url).text
 
-    @cached(CACHE)
     def get_content_raw(self, url: str) -> bytes:
         return self._session.get(url).content
 
